<?php
/**
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 * @license MIT
 */

declare(strict_types=1);

namespace FeedReaderCentral;

use BitBadger\InspiredByFSharp\{Option, Result};
use BitBadger\PDODocument\{
    Configuration, Custom, Document, DocumentException, DocumentList, Exists, Field, Find, Parameters, Patch, Query
};
use DateTimeInterface;

/**
 * An RSS or Atom feed
 */
class Feed
{
    // ***** CONSTANTS *****

    /** @var int Do not purge items */
    public const PurgeNone = 0;

    /** @var int Purge all read items (will not purge unread items) */
    public const PurgeRead = 1;

    /** @var int Purge items older than the specified number of days */
    public const PurgeByDays = 2;

    /** @var int Purge items in number greater than the specified number of items to keep */
    public const PurgeByCount = 3;

    /**
     * @var int Do not purge items
     * @deprecated Use Feed::PurgeNone instead
     */
    public const PURGE_NONE = 0;

    /**
     * @var int Purge all read items (will not purge unread items)
     * @deprecated Use Feed::PurgeRead instead
     */
    public const PURGE_READ = 1;

    /**
     * @var int Purge items older than the specified number of days
     * @deprecated Use Feed::PurgeByDays instead
     */
    public const PURGE_BY_DAYS = 2;

    /**
     * @var int Purge items in number greater than the specified number of items to keep
     * @deprecated Use Feed::PurgeByCount instead
     */
    public const PURGE_BY_COUNT = 3;

    /**
     * Constructor
     *
     * @param int $id The ID of the feed
     * @param int $user_id The ID of the user to whom this subscription belongs
     * @param string $url The URL of the feed
     * @param string|null $title The title of this feed
     * @param string|null $updated_on The date/time items in this feed were last updated
     * @param string|null $checked_on The date/time this feed was last checked
     */
    public function __construct(public int $id = 0, public int $user_id = 0, public string $url = '',
                                public ?string $title = null, public ?string $updated_on = null,
                                public ?string $checked_on = null) { }

    // ***** STATIC FUNCTIONS *****

    /**
     * Create a document from the parsed feed
     *
     * @param ParsedFeed $parsed The parsed feed
     * @return Feed The document constructed from the parsed feed
     */
    public static function fromParsed(ParsedFeed $parsed): self
    {
        return new self(
            user_id:    $_SESSION[Key::UserId],
            url:        $parsed->url,
            title:      $parsed->title,
            updated_on: $parsed->updatedOn,
            checked_on: Data::formatDate('now'));
    }

    /**
     * Update a feed's items
     *
     * @param int $feedId The ID of the feed to which these items belong
     * @param ParsedFeed $parsed The extracted Atom or RSS feed items
     * @param DateTimeInterface $lastChecked When this feed was last checked (only new items will be added)
     * @return Result<true, string> True if successful, an error message if not
     */
    public static function updateItems(int $feedId, ParsedFeed $parsed, DateTimeInterface $lastChecked): Result
    {
        $results =
            array_map(function ($item) use ($feedId) {
                try {
                    $tryExisting = Find::firstByFields(Table::Item,
                        [Field::EQ('feed_id', $feedId), Field::EQ('item_guid', $item->guid)], Item::class);
                    if ($tryExisting->isSome()) {
                        $existing = $tryExisting->get();
                        if ($existing->published_on !== $item->publishedOn
                            || ($existing->updated_on !== ($item->updatedOn ?? ''))) {
                            Item::update($existing->id, $item);
                        }
                    } else {
                        Item::add($feedId, $item);
                    }
                    return Result::OK(true);
                } catch (DocumentException $ex) {
                    return Result::Error("$ex");
                }
            }, array_filter($parsed->items,
                fn(ParsedItem $it) => date_create_immutable($it->updatedOn ?? $it->publishedOn) >= $lastChecked));
        $errors = array_map(fn(Result $it) => $it->getError(), array_filter($results, fn($it) => $it->isError()));
        return sizeof($errors) > 0 ? Result::Error(implode("\n", $errors)) : Result::OK(true);
    }

    /**
     * Purge items for a feed
     *
     * @param int $feedId The ID of the feed to be purged
     * @return Result<true, string> True if purging was successful, an error message if not
     * @throws DocumentException If any is encountered
     */
    private static function purgeItems(int $feedId): Result
    {
        $fields = [Field::EQ('feed_id', $feedId, ':feed'), Data::bookmarkField(false)];
        $sql    = Query\Delete::byFields(Table::Item, $fields);

        switch (PURGE_TYPE) {
            case self::PurgeRead:
                $readField = Field::EQ('is_read', 1, ':read');
                $fields[]  = $readField;
                $sql      .= ' AND ' . Query::whereByFields([$readField]);
                break;
            case self::PurgeByDays:
                $fields[] = Field::EQ('', Data::formatDate('-' . PURGE_NUMBER . ' day'), ':oldest');
                $sql     .= " AND date(coalesce(data->>'updated_on', data->>'published_on')) < date(:oldest)";
                break;
            case self::PurgeByCount:
                $fields[] = Field::EQ('', PURGE_NUMBER, ':keep');
                $id       = Configuration::$idField;
                $table    = Table::Item;
                $sql     .= ' ' . <<<SQL
                    AND data->>'$id' IN (
                        SELECT data->>'$id' FROM $table
                         WHERE data->>'feed_id' = :feed
                         ORDER BY date(coalesce(data->>'updated_on', data->>'published_on')) DESC
                         LIMIT -1 OFFSET :keep
                    )
                    SQL;
                break;
            default:
                return Result::Error('Unrecognized purge type ' . PURGE_TYPE);
        }
        try {
            Custom::nonQuery($sql, Parameters::addFields($fields, []));
            return Result::OK(true);
        } catch (DocumentException $ex) {
            return Result::Error("$ex");
        }
    }

    /**
     * Refresh a feed
     *
     * @param int $feedId The ID of the feed to be refreshed
     * @param string $url The URL of the feed to be refreshed
     * @return Result<true, string> True if successful, an error message if not
     */
    public static function refreshFeed(int $feedId, string $url): Result
    {
        return ParsedFeed::retrieve($url)
            ->bind(function (ParsedFeed $feed) use ($feedId, $url) {
                try {
                    $feedDoc = Find::byId(Table::Feed, $feedId, self::class);
                    if ($feedDoc->isNone()) return Result::Error('Could not derive date last checked for feed');
                    $lastChecked = date_create_immutable($feedDoc->get()->checked_on ?? WWW_EPOCH);

                    return self::updateItems($feedId, $feed, $lastChecked)
                        ->bind(function () use ($feed, $feedId, $url) {
                            $patch = [
                                'title'      => $feed->title,
                                'updated_on' => $feed->updatedOn,
                                'checked_on' => Data::formatDate('now')
                            ];
                            if ($url !== $feed->url) $patch['url'] = $feed->url;
                            Patch::byId(Table::Feed, $feedId, $patch);
                            return Result::OK(true);
                        })
                        ->bind(fn() => PURGE_TYPE === self::PurgeNone ? Result::OK(true) : self::purgeItems($feedId));
                } catch (DocumentException $ex) {
                    return Result::Error("$ex");
                }
            });
    }

    /**
     * Add an RSS feed
     *
     * @param string $url The URL of the RSS feed to add
     * @return Result<int, string> The feed ID if successful, an error message if not
     */
    public static function add(string $url): Result
    {
        return ParsedFeed::retrieve($url)
            ->bind(function (ParsedFeed $feed) {
                try {
                    $fields = [Field::EQ('user_id', $_SESSION[Key::UserId]), Field::EQ('url', $feed->url)];
                    if (Exists::byFields(Table::Feed, $fields)) {
                        return Result::Error("Already subscribed to feed $feed->url");
                    }

                    Document::insert(Table::Feed, self::fromParsed($feed));

                    $doc = Find::firstByFields(Table::Feed, $fields, self::class);
                    if ($doc->isNone()) return Result::Error('Could not retrieve inserted feed');

                    return self::updateItems($doc->get()->id, $feed, date_create_immutable(WWW_EPOCH))
                        ->bind(fn() => Result::OK($doc->get()->id));
                } catch (DocumentException $ex) {
                    return Result::Error("$ex");
                }
            });
    }

    /**
     * Update an RSS feed
     *
     * @param Feed $existing The existing feed
     * @param string $url The URL with which the existing feed should be modified
     * @return Result<true, string> True if successful, an error message if not
     */
    public static function update(Feed $existing, string $url): Result
    {
        try {
            Patch::byFields(Table::Feed,
                [Field::EQ(Configuration::$idField, $existing->id), Field::EQ('user_id', $_SESSION[Key::UserId])],
                ['url' => $url]);
            return self::refreshFeed($existing->id, $url);
        } catch (DocumentException $ex) {
            return Result::Error("$ex");
        }
    }

    /**
     * Retrieve all feeds, optionally for a specific user
     *
     * @param int $user The ID of the user whose feeds should be retrieved (optional, defaults to all feeds)
     * @return DocumentList<Feed> A list of feeds
     * @throws DocumentException If any is encountered
     */
    public static function retrieveAll(int $user = 0): DocumentList
    {
        return $user === 0
            ? Find::all(Table::Feed, self::class)
            : Find::byFields(Table::Feed, [Field::EQ('user_id', $user)], self::class);
    }

    /**
     * Refresh all feeds
     *
     * @return Result<true, string> True if successful an error message if not (may have multiple error lines)
     */
    public static function refreshAll(): Result
    {
        $errors = [];

        try {
            self::retrieveAll($_SESSION[Key::UserId])->iter(function (Feed $feed) use (&$errors) {
                self::refreshFeed($feed->id, $feed->url)
                    ->mapError(function (string $err) use (&$errors) { $errors[] = $err; });
            });
        } catch (DocumentException $ex) {
            return Result::Error("$ex");
        }

        return empty($errors) ? Result::OK(true) : Result::Error(implode("\n", $errors));
    }

    /**
     * Retrieve a feed by its ID for the current user
     *
     * @param int $feedId The ID of the feed to retrieve
     * @return Option<Feed> A `Some` value with the data for the feed if found, `None` otherwise
     * @throws DocumentException If any is encountered
     */
    public static function retrieveById(int $feedId): Option
    {
        return Find::byId(Table::Feed, $feedId, static::class)
            ->filter(fn($it) => $it->user_id === $_SESSION[Key::UserId]);
    }
}
