<?php
/**
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 * @license MIT
 */

declare(strict_types=1);

namespace FeedReaderCentral;

use BitBadger\PDODocument\{Document, DocumentException, Patch};

/**
 * An item from a feed
 */
class Item
{
    /**
     * Constructor
     *
     * @param int $id The ID of this item in the Feed Reader Central database
     * @param int $feed_id The ID of the feed to which this item belongs
     * @param string $title The title of this item
     * @param string $item_guid The Globally Unique ID (GUID) for this item (an attribute in the feed XML)
     * @param string $item_link The link to the item on its original site
     * @param string $published_on The date/time this item was published
     * @param string|null $updated_on The date/time this item was last updated
     * @param string $content The content for this item
     * @param int $is_read 1 if the item has been read, 0 if not
     * @param int $is_bookmarked 1 if the item is bookmarked, 0 if not
     */
    public function __construct(public int $id = 0, public int $feed_id = 0, public string $title = '',
                                public string $item_guid = '', public string $item_link = '',
                                public string $published_on = '', public ?string $updated_on = null,
                                public string $content = '', public int $is_read = 0, public int $is_bookmarked = 0) { }

    /**
     * Has the item been read?
     *
     * @return bool True if the item has been read, false if not
     */
    public function isRead(): bool
    {
        return $this->is_read <> 0;
    }

    /**
     * Is the item bookmarked?
     *
     * @return bool True if the item is bookmarked, false if not
     */
    public function isBookmarked(): bool
    {
        return $this->is_bookmarked <> 0;
    }

    /**
     * Add an item
     *
     * @param int $feedId The ID of the feed to which the item belongs
     * @param ParsedItem $parsed The parsed item from the feed XML
     * @throws DocumentException If any is encountered
     */
    public static function add(int $feedId, ParsedItem $parsed): void
    {
        Document::insert(Table::Item, new self(
            feed_id:      $feedId,
            title:        $parsed->title,
            item_guid:    $parsed->guid,
            item_link:    $parsed->link,
            published_on: $parsed->publishedOn,
            updated_on:   $parsed->updatedOn,
            content:      $parsed->content));
    }

    /**
     * Update an item
     *
     * @param int $id The ID of the item to be updated
     * @param ParsedItem $parsed The parsed item from the feed XML
     * @throws DocumentException If any is encountered
     */
    public static function update(int $id, ParsedItem $parsed): void
    {
        Patch::byId(Table::Item, $id, [
            'title'        => $parsed->title,
            'published_on' => $parsed->publishedOn,
            'updated_on'   => $parsed->updatedOn,
            'content'      => $parsed->content,
            'is_read'      => 0
        ]);
    }
}
