<?php
/**
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 * @license MIT
 */

declare(strict_types=1);

namespace FeedReaderCentral;

use BitBadger\InspiredByFSharp\Option;
use BitBadger\PDODocument\{Configuration, Custom, DocumentException, DocumentList, Field, Parameters, Query};
use BitBadger\PDODocument\Mapper\DocumentMapper;

/**
 * A list of items to be displayed
 *
 * This is a wrapper for retrieval and display of arbitrary lists of items based on a SQLite result.
 */
class ItemList
{
    /** @var DocumentList<ItemWithFeed> The items matching the criteria, lazily iterable */
    private DocumentList $dbList;

    /** @var Option<string> The error message generated by executing a query */
    private Option $error;

    /** @var bool Whether to render a link to the feed to which the item belongs */
    public bool $linkFeed = false;

    /** @var bool Whether to display the feed to which the item belongs */
    public bool $displayFeed = false;

    /** @var bool Whether to show read / bookmarked indicators on posts */
    public bool $showIndicators = false;

    /**
     * Constructor
     *
     * @param string $itemType The type of item being displayed (unread, bookmark, etc.)
     * @param string $returnURL The URL to which the item page should return once the item has been viewed
     * @param Field[] $fields The fields to use to restrict the results
     * @param string $searchWhere Additional WHERE clause to use for searching
     */
    private function __construct(public string $itemType, public string $returnURL = '', array $fields = [],
                                 string $searchWhere = '')
    {
        $this->error = Option::None();
        $allFields   = [Data::userIdField(Table::Feed), ...$fields];
        try {
            $this->dbList = Custom::list(
                ItemWithFeed::SELECT_WITH_FEED . ' WHERE '
                    . Query::whereByFields(array_filter($allFields, fn($it) => $it->paramName !== ':search'))
                . "$searchWhere ORDER BY coalesce(item.data->>'updated_on', item.data->>'published_on') DESC",
                Parameters::addFields($allFields, []), new DocumentMapper(ItemWithFeed::class));
        } catch (DocumentException $ex) {
            $this->error = Option::Some("$ex");
        }
    }

    /**
     * Create an item list with all the current user's bookmarked items
     *
     * @return ItemList An item list with all bookmarked items
     */
    public static function allBookmarked(): self
    {
        $list = new self('Bookmarked', '/?bookmarked', [Data::bookmarkField(qualifier: Table::Item)]);
        $list->linkFeed = true;
        return $list;
    }

    /**
     * Create an item list with all the current user's unread items
     *
     * @return ItemList An item list with all unread items
     */
    public static function allUnread(): self
    {
        $list = new self('Unread', fields: [Data::unreadField(Table::Item)]);
        $list->linkFeed = true;
        return $list;
    }

    /**
     * Create an item list with all items for the given feed
     *
     * @param int $feedId The ID of the feed for which items should be retrieved
     * @return ItemList An item list with all items for the given feed
     */
    public static function allForFeed(int $feedId): self
    {
        $list = new self('', "/feed/items?id=$feedId", [Data::feedField($feedId, Table::Feed)]);
        $list->showIndicators = true;
        return $list;
    }

    /**
     * Create an item list with unread items for the given feed
     *
     * @param int $feedId The ID of the feed for which items should be retrieved
     * @return ItemList An item list with unread items for the given feed
     */
    public static function unreadForFeed(int $feedId): self
    {
        return new self('Unread', "/feed/items?id=$feedId&unread",
            [Data::feedField($feedId, Table::Feed), Data::unreadField(Table::Item)]);
    }

    /**
     * Create an item list with bookmarked items for the given feed
     *
     * @param int $feedId The ID of the feed for which items should be retrieved
     * @return ItemList An item list with bookmarked items for the given feed
     */
    public static function bookmarkedForFeed(int $feedId): self
    {
        return new self('Bookmarked', "/feed/items?id=$feedId&bookmarked",
            [Data::feedField($feedId, Table::Feed), Data::bookmarkField(qualifier: Table::Item)]);
    }

    /**
     * Create an item list with items matching given search terms
     *
     * @param string $search The item search terms / query
     * @param bool $isBookmarked Whether to restrict the search to bookmarked items
     * @return ItemList An item list match the given search terms
     */
    public static function matchingSearch(string $search, bool $isBookmarked): self
    {
        $fields = [Field::EQ('content', $search, ':search')];
        if ($isBookmarked) $fields[] = Data::bookmarkField(qualifier: Table::Item);
        $list = new self('Matching' . ($isBookmarked ? ' Bookmarked' : ''),
            "/search?search=$search&items=" . ($isBookmarked ? 'bookmarked' : 'all'), $fields,
            ' AND ' . Table::Item . ".data->>'" . Configuration::$idField . "' IN "
                . '(SELECT ROWID FROM item_search WHERE content MATCH :search)');
        $list->showIndicators = true;
        $list->displayFeed    = true;
        return $list;
    }

    /**
     * Render this item list
     */
    public function render(): void
    {
        if ($this->error->isSome()) {
            echo "<p>Error retrieving list:<br>{$this->error->get()}";
            return;
        }
        $return = $this->returnURL === '' ? '' : '&from=' . urlencode($this->returnURL);
        echo '<article>';
        if ($this->dbList->hasItems()) {
            $this->dbList->iter(function (ItemWithFeed $item) use ($return) {
                echo '<p>' . hx_get("/item?id=$item->id$return", strip_tags($item->title)) . '<br><small>';
                if ($this->showIndicators) {
                    if (!$item->isRead()) echo '<strong>Unread</strong> &nbsp; ';
                    if ($item->isBookmarked()) echo '<strong>Bookmarked</strong> &nbsp; ';
                }
                echo '<em>' . date_time($item->updated_on ?? $item->published_on) . '</em>';
                if ($this->linkFeed) {
                    echo ' &bull; ' .
                        hx_get("/feed/items?id={$item->feed->id}&" . strtolower($this->itemType),
                            htmlentities($item->feed->title));
                } elseif ($this->displayFeed) {
                    echo ' &bull; ' . htmlentities($item->feed->title);
                }
                echo '</small>';
            });
        } else {
            echo '<p><em>There are no ' . strtolower($this->itemType) . ' items</em>';
        }
        echo '</article>';
    }
}
