<?php
/**
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 * @license MIT
 */

declare(strict_types=1);

namespace FeedReaderCentral;

use DOMNode;

/**
 * Information for a feed item
 */
readonly class ParsedItem
{
    /**
     * Constructor
     *
     * @param string $guid The unique ID for the feed item
     * @param string $title The title of the feed item
     * @param string $link The link to the original content
     * @param string $publishedOn When this item was published
     * @param string|null $updatedOn When this item was last updated
     * @param string $content The content for the item
     */
    private function __construct(public string $guid = '', public string $title = '', public string $link = '',
                                 public string $publishedOn = '', public ?string $updatedOn = null,
                                 public string $content = '') { }

    /**
     * Construct a feed item from an Atom feed's `<entry>` tag
     *
     * @param DOMNode $node The XML node from which a feed item should be constructed
     * @return ParsedItem A feed item constructed from the given node
     */
    public static function fromAtom(DOMNode $node): self
    {
        $guid = ParsedFeed::atomValue($node, 'id');
        $link = '';
        foreach ($node->getElementsByTagName('link') as $linkElt) {
            if ($linkElt->hasAttributes()) {
                $relAttr = $linkElt->attributes->getNamedItem('rel');
                if ($relAttr && $relAttr->value === 'alternate') {
                    $link = $linkElt->attributes->getNamedItem('href')->value;
                    break;
                }
            }
        }
        if ($link === '' && str_starts_with($guid, 'http')) $link = $guid;

        return new self(
            guid:        $guid,
            title:       ParsedFeed::atomValue($node, 'title'),
            link:        $link,
            publishedOn: Data::formatDate(ParsedFeed::atomValue($node, 'published')),
            updatedOn:   Data::formatDate(ParsedFeed::atomValue($node, 'updated')),
            content:     ParsedFeed::atomValue($node, 'content'));
    }

    /**
     * Construct a feed item from an RSS feed's `<item>` tag
     *
     * @param DOMNode $node The XML node from which a feed item should be constructed
     * @return ParsedItem A feed item constructed from the given node
     */
    public static function fromRSS(DOMNode $node): self
    {
        $itemGuid = ParsedFeed::rssValue($node, 'guid');
        $updNodes = $node->getElementsByTagNameNS(ParsedFeed::ATOM_NS,    'updated');
        $encNodes = $node->getElementsByTagNameNS(ParsedFeed::CONTENT_NS, 'encoded');

        return new self(
            guid:        $itemGuid == 'guid not found' ? ParsedFeed::rssValue($node, 'link') : $itemGuid,
            title:       ParsedFeed::rssValue($node, 'title'),
            link:        ParsedFeed::rssValue($node, 'link'),
            publishedOn: Data::formatDate(ParsedFeed::rssValue($node, 'pubDate')),
            updatedOn:   Data::formatDate($updNodes->length > 0 ? $updNodes->item(0)->textContent : null),
            content:     $encNodes->length > 0
                ? $encNodes->item(0)->textContent
                : ParsedFeed::rssValue($node, 'description'));
    }
}
