<?php
/**
 * Alpha 7 -> Beta 1 Database Update Utility
 *
 * Between these two versions, the data format changed; this utility migrates existing data to its new format.
 *
 * _It will be removed in v1.1._
 *
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 * @license MIT
 */

declare(strict_types=1);

use BitBadger\PDODocument\{Custom, Document, DocumentException};
use BitBadger\PDODocument\Mapper\{ArrayMapper, ExistsMapper};
use FeedReaderCentral\{Data, Feed, Item, Table, User};

require __DIR__ . '/../cli-start.php';

cli_title('DATABASE UPDATE');

if ($argc < 2) display_help();

switch ($argv[1]) {
    case 'check':
        check_status();
        break;
    case 'run':
        run_update();
        break;
    default:
        printfn('Unrecognized option "%s"', $argv[1]);
        display_help();
}


/**
 * Display the options for this utility and exit
 */
function display_help(): never
{
    printfn('Options:');
    printfn(' - check');
    printfn('     Check to see if the configured database has been updated');
    printfn(' - run');
    printfn('     Run the beta1 database storage update');
    exit(0);
}

function json_column_exists(): bool
{
    try {
        $table = Custom::single("SELECT sql FROM sqlite_master WHERE tbl_name='frc_user'", [], new ArrayMapper())
            ->getOrDefault(['sql' => '']);
        return $table && substr_compare(strtolower($table['sql']), 'data text not null', 0) >= 0;
    } catch (DocumentException $ex) {
        printfn("ERR $ex");
        return false;
    }
}

function check_status(): void
{
    if (json_column_exists()) {
        printfn('The database has already been updated');
    } else {
        printfn('The database has yet to be updated');
    }
}

function run_update(): void
{
    try {
        $searchExists = Custom::scalar("SELECT EXISTS (SELECT 1 FROM sqlite_master WHERE name = 'item_ai')", [],
            new ExistsMapper());
        if ($searchExists) {
            printfn('Removing search index...');
            Custom::nonQuery('DROP TRIGGER item_ai', []);
            Custom::nonQuery('DROP TRIGGER item_au', []);
            Custom::nonQuery('DROP TRIGGER item_ad', []);
            Custom::nonQuery('DROP TABLE item_search', []);
        }
        printfn('Moving old tables...');
        Custom::nonQuery('ALTER TABLE item     RENAME TO old_item', []);
        Custom::nonQuery('ALTER TABLE feed     RENAME TO old_feed', []);
        Custom::nonQuery('ALTER TABLE frc_user RENAME TO old_user', []);
        printfn('Creating new tables...');
        Data::ensureDb();
        printfn('Migrating users...');
        $users = Custom::list('SELECT * FROM old_user', [], new ArrayMapper());
        if (!$users->hasItems()) throw new DocumentException('Could not retrieve users');
        foreach ($users->items() as $user) {
            Document::insert(Table::User, new User($user['id'], $user['email'], $user['password']));
        }
        printfn('Migrating feeds...');
        $feeds = Custom::list('SELECT * FROM old_feed', [], new ArrayMapper());
        if (!$feeds->hasItems()) throw new DocumentException('Could not retrieve feeds');
        foreach ($feeds->items() as $feed) {
            Document::insert(Table::Feed,
                new Feed($feed['id'], $feed['user_id'], $feed['url'], $feed['title'], $feed['updated_on'],
                    $feed['checked_on']));
        }
        printfn('Migrating items...');
        $items = Custom::list('SELECT * FROM old_item', [], new ArrayMapper());
        if (!$items->hasItems()) throw new DocumentException('Could not retrieve items');
        foreach ($items->items() as $item) {
            Document::insert(Table::Item,
                new Item($item['id'], $item['feed_id'], $item['title'], $item['item_guid'], $item['item_link'],
                    $item['published_on'], $item['updated_on'], $item['content'], $item['is_read'],
                    $item['is_bookmarked']));
        }
        printfn('Dropping old tables...');
        Custom::nonQuery('DROP TABLE old_item', []);
        Custom::nonQuery('DROP TABLE old_feed', []);
        Custom::nonQuery('DROP TABLE old_user', []);
        printfn(PHP_EOL. 'Migration complete!');
    } catch (DocumentException $ex) {
        printfn("ERR $ex");
    }
}
